// prcsPointCloudFile.cpp
// The utility processes 3D scan files saved by the ROS-based CTU Data Logger utility running on NIFTi robot.
// The input files are to be provided as launch parameters (which are the only parameters expected).
// This file contains the whole code.
// It is intended for Win32 as well as for Linux
// Depending on the system, some parts has to be (un)commented:

// WIN32 only
#include "stdafx.h"
#include <Windows.h>
#define SLEEP Sleep(2000)


// LINUX only
//#include <unistd.h>
//#include <memory.h>
//#define SLEEP sleep(2)




// common
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <string>
#include <stdlib.h>

using namespace std;



int main(int argc, char* argv[])

{
	
	// Variables definition 

	const double CROP_RADIUS_SQR = 0.2; // This sets a radius around the 3D scan frame origin. 
										//In this radius, points can be cropped out (depends on the user choise) 
	bool cropRobot;			// bool setting the crop process
	int i;					// iterators
	int k;
	int u;
	int fileIter;

	
	int progressFract;		//contains a single file process progress
	char* inFileName;		// filenames and file handlers
	ifstream* inputFile;
	ofstream* outputFileASC;
	ofstream* outputFilePCD;
	string ident;
	string fNameASC;
	string fNamePCD;

	int temp;				// temps for point counting and header processing
	long int temp1;
	long int temp2;
	
	long int points;
	long int finalPoints;
	char sbuff[101];
	char tempChar;

	int coordPositions[3];		// contains coordinates (order in a single raw file line) of the three axes saved in the 3D scan file
	int* coordPositionsInv;		// contains the same information but inverted: if 5 numbers saved and x~2,y~3,z~4, than this is 01110

	int* varTypes;				// numbers identifiing number formats in the raw file
	int* varLengths;			// each number byte count
	int varCount;				// number of numbers in a single line

	float XYZ[3];				// contains a single point coordinates extracted from the raw file


	unsigned char bytes[8];     // is used to contruct float64 from 8bytes or more likely float32 from 4 bytes from the raw file  
	float coord;				// the output float32

	// Check if any parameters entered, if not, finish - no files to process

	if (argc == 1) {
		cout << "Will not start w/o parameter(s): file1Path [file2Path] [file3Path] ... " << endl << "Finished" << endl;
		SLEEP;
		return 0;
	} 

	// ask if user wants to crop points near to the origin
	cropRobot = true;
	cout << "To keep pointclouds as they are, press 'n' and Enter."<< endl
		 << "To leave out points inside sphere of x^2+y^2+z^2=" << CROP_RADIUS_SQR <<" [m]; press anything else and Enter" << endl << ":";

	cin >> ident;
	if(!ident.compare("n")){
		cropRobot = false;
	}

	ident.clear();



	// this is the main loop going through all parametrs, loading input files and processing them

	for(fileIter = 1; fileIter < argc; fileIter++){
		
		// variables initializitaion / reinitialization

		i = 0;
		k = 0;
		u = 0;
		temp = 0;
		progressFract = 0;
		inFileName = 0; 
		inputFile = 0;
		outputFileASC = 0;
		outputFilePCD = 0;
		temp1 = 0;
		temp2 = 0;
		points = 0;
		finalPoints = 0;
		tempChar = 0;
		coordPositions[0] = -1; coordPositions[1] = -1; coordPositions[2] = -1;
		coordPositionsInv = 0;

		varTypes = 0;
		varLengths = 0;
		varCount = 0;

		coord = 0;
		ident.clear();
		fNameASC.clear();
		fNamePCD.clear();


		// loading appropriate input parameter (filename)


		inFileName = argv[fileIter];   // the input filename
		ident = argv[fileIter];			// the output filename (will be the same except its suffix)

		// delete suffix from the output filename
		i = ident.rfind('.');
		if(i != ident.npos){
			ident.erase(i);
		}

		cout << "CTU Logger PointCloud File: " << inFileName << endl << endl;
		cout << "Identificator of the files to be created: " << ident << endl << endl;
		
		
	
		// open input file
		inputFile = new ifstream();
		inputFile->open(inFileName);

		if(!inputFile->is_open()){
			cout << "File " << inFileName << " can not be opened, skipping." << endl;
			delete inputFile;
			continue;

		}else{
			cout << "File " << inFileName << " opened, creating \"" << ident << ".asc and "<< ident << ".pcd" << endl;
		}

		// create names for two new files
		fNameASC.append(ident);
		fNameASC.append(".asc");

		fNamePCD.append(ident);
		fNamePCD.append(".pcd");


		// create the two new files
		outputFileASC = new ofstream(fNameASC.c_str());
		outputFilePCD = new ofstream(fNamePCD.c_str());

		if(! (outputFileASC->is_open() && outputFilePCD->is_open() )){
			
			cout << endl << "ERROR:at least one of the files to write to could not be created."<< endl;
			cout << "Check if you have sufficient rights to create files..." << endl;

			inputFile->close();
			delete inputFile;
			
			delete outputFileASC;
			delete outputFilePCD;

			SLEEP;
			SLEEP;
			return 0;
	
		}
		// HEADER PROCESSING
		// read the input file header
			
		// read the first newline
		inputFile->getline(sbuff,100);
		// read the timestamp line
		inputFile->getline(sbuff,100);

		sscanf(sbuff,"%d.%d",&temp1,&temp2);
	
		cout << "TimeStamp1 =" << temp1 << endl;
		cout << "TimeStamp2 =" << temp2 << endl;
	
		// read height and width
		inputFile->getline(sbuff,100);

		sscanf(sbuff,"%d, %d",&temp1,&points);

		// read number of coordinates
		inputFile->getline(sbuff,100);
		sscanf(sbuff,"%d",&temp1);
		varCount = temp1;

		if(varCount < 3){
			cout << "Error, point cloud file incompatible (not enough variables)";
			inputFile->close();
			delete inputFile;
			outputFileASC->close();
			delete outputFileASC;
			outputFilePCD->close();
			delete outputFilePCD;

			SLEEP;
			return 0;
		}


		// there are temp1 coordinates, lets find xyz
		
		// alloc space for appropriate number of numbers present in a sigle raw file line
		varTypes = new int[varCount];
		varLengths = new int[varCount];
		coordPositionsInv = new int[varCount];

		// going through a header part listing all numbers saved in a single line (likely: X Y Z Intenzity Order)
		for(i = 0; i < varCount; i ++){
	
			inputFile->getline(sbuff,100);

			// notice where the interesting coordinates are
			if(sbuff[0] == 'x') coordPositions[0] = i; 
			if(sbuff[0] == 'y') coordPositions[1] = i;
			if(sbuff[0] == 'z') coordPositions[2] = i;

			// skipping the next line
			inputFile->getline(sbuff,100);
			
			// read the variable type
			inputFile->getline(sbuff,100);

			// store the type of each variable
			sscanf(sbuff,"%d",varTypes+i);

			// skip the last line
			inputFile->getline(sbuff,100);
	
		}

		// save appropriate variable lengths in bytes according to the PointCloud2 specification
		/*	
			uint8 INT8=1
			uint8 UINT8=2
			uint8 INT16=3
			uint8 UINT16=4
			uint8 INT32=5
			uint8 UINT32=6
			uint8 FLOAT32=7
			uint8 FLOAT64=8
		*/
		for(i = 0; i < varCount; i++){
		
			switch(varTypes[i]){
		
			case 1:
			case 2:
				varLengths[i] = 1;
				break;
			case 3:
			case 4:
				varLengths[i] = 2;
				break;		
			case 5:
			case 6:
			case 7:
				varLengths[i] = 4;
				break;
			case 8:
				varLengths[i] = 8;
				break;
			}
		

		}

		// check if all x,y,z coordinates are present in the file (were searched for automatically so it might happen that not all are present)
		if(coordPositions[0] == -1 || coordPositions[1] == -1 || coordPositions[1] == -1){
			cout << "Error, at least one coordinate of x, y or z is missing.";
			inputFile->close();
			delete inputFile;
			outputFileASC->close();
			delete outputFileASC;
			outputFilePCD->close();
			delete outputFilePCD;
			delete varTypes;
			delete varLengths;
			delete coordPositionsInv;

			SLEEP;
			return 0;
		}

		
		// fill the coordPositionsInv with positions of x,y,z, rest -1;
		for(i = 0; i < varCount; i++) coordPositionsInv[i] = -1;
		for(i = 0; i < 3; i++) coordPositionsInv[coordPositions[i]] = i;
	

		// skip the rest of lines of the header
		inputFile->getline(sbuff,100);
		inputFile->getline(sbuff,100);
	
		
		// FILE PROCESSING
		
		// Depending on the user crop choise, print the file information

		if(cropRobot){
		
			cout << "Point cloud contains " << points << " to be extrated, robot points within sqrt(" <<CROP_RADIUS_SQR<<") radius will be cropped" << endl;
	
		}else{
			cout << "Point cloud contains " << points << " to be extrated (no cropping)" << endl;
		}

		// skip lines under the header
		inputFile->getline(sbuff,100);
		inputFile->getline(sbuff,100);
	
		
		cout << "Progress of .asc file %:" << endl;
	
	
      
		// process the rest of the input file and write the float values
		
		
		progressFract = points/10;  // for progress printing

		// cycle for all points in the 3D scan
		for(k = 0; k < points; k++){

			// print the progress
			if((k % progressFract) == 0)
				cout << (k/progressFract)*10 << " ";

			// cycle through a whole line of byte values
			for(u = 0; u < varCount; u++){
				
				temp = 0;


				// read appropriate number of bytes for each number in the line 
				// and based on coordPositionsInv decide wheather to save or discard them
				if(coordPositionsInv[u] != -1) {  

					// this variable is one of x,y,z, we expect float32
					for(i = 0; i < varLengths[u] ; i ++){   // read the bytes (saved in ascii)
						inputFile->get(sbuff,100,' '); // read the byte
						inputFile->get();			// skip the separator
						sscanf(sbuff,"%d",&temp);	// process the byte: ascii -> binary
						bytes[i] = (unsigned char) temp; // put to its place in the float
					}

					// construct float
					// it is expected, that xyz coordinates are saved as float32. this will construct jibberish otherwise
					memcpy(&coord, &bytes, sizeof(coord));

					// store for further writing
					XYZ[coordPositionsInv[u]] = coord;
			
				}else{
					// this must some other variable, which doesnt interest us
					// we'll just read the bytes
					for(i = 0; i < varLengths[u] ; i ++){
						inputFile->get(sbuff,100,' ');
						inputFile->get();
					}
				}
			

			}
			
				// check if user wants to crop robot from the cloud (default = yes)
				// write XYZ and newline
				// write to the .asc file. The .pcd file need the point count in its header. It is not known since some of the points
				// may be cropped; the .pcd file will be created later.

				if(cropRobot){
				
					//all inside ball of one meter radius around origin wont be included into cloud
					if((XYZ[0]*XYZ[0] + XYZ[1]*XYZ[1] + XYZ[2]*XYZ[2]) > CROP_RADIUS_SQR){
						(*outputFileASC) << XYZ[0] << " " << XYZ[1] << " " << XYZ[2] << "\n";
						finalPoints++;
					}


				}else{
					
					(*outputFileASC) << XYZ[0] << " " << XYZ[1] << " " << XYZ[2] << "\n";
					finalPoints++;
				}

		}

		// input file processed, the .asc file contains all the points.
		// finalPoints contains number of all points saved to the .asc file.

		// we will create the .pcd file by copying all the points from the .asc file
		cout << "Copying data to .pcd file..." << endl;


		// PCD file header preparation
		(*outputFilePCD) << "# .PCD v.7 - Point Cloud Data file format\n";
		(*outputFilePCD) << "VERSION .7\n";
		(*outputFilePCD) << "FIELDS x y z\n";
		(*outputFilePCD) << "SIZE 4 4 4\n";
		(*outputFilePCD) << "TYPE F F F\n";
		(*outputFilePCD) << "COUNT 1 1 1\n";
		(*outputFilePCD) << "WIDTH " << finalPoints << "\n";
		(*outputFilePCD) << "HEIGHT 1\n";
		(*outputFilePCD) << "VIEWPOINT 0 0 0 1 0 0 0\n";
		(*outputFilePCD) << "POINTS " << finalPoints << "\n";
		(*outputFilePCD) << "DATA ascii\n";


		// closing input file and asc file(for writing)
		outputFileASC->close();
		delete outputFileASC;

		inputFile->close();
		delete inputFile;


		// open asc file for reading
		inputFile = new ifstream();
		inputFile->open(fNameASC.c_str());

		//copy whole asc to pcd file (under its header)

		while(1){
		
			tempChar = (char)(inputFile->get());
			if(inputFile->eof()) break;
			(*outputFilePCD) << tempChar;
	
		}

	
		cout << "Done." << endl;

	
		// close all file and dealloc all memory.

		inputFile->close();
		delete inputFile;

		outputFilePCD->close();
		delete outputFilePCD;

		delete varTypes;
		delete varLengths;
		delete coordPositionsInv;

		// continue to the next input file or end
	}
		
	SLEEP;
	return 0;
}
