% psg = tzgui(ps,parname,gsz,xname,fname) - GUI for experiments
% 
% ps        ... structure with ps.ParName1, ps.ParName2, ... 
% parname   ... the name of the structure as a string
% gsz       ... the size of the grid for buttons [rnum cnum] 
% xname     ... callback function name
% fname     ... callback button label
% psg       ... GUI info structure
%
% The function creates a window with buttons for each field in ps named by
% the respective field name. The type of the button is determined by the
% type of the variable:
%                       logical ... radio button
%                       other   ... edit field
% Editing the fields changes status of ps. Pressing the single execution
% button calls function xname.
%
% See also TZ2011, CLEARALL, 

% T. Pajdla, pajdla@cmp.felk.cvut.cz
% 14 Feb 2011
function psg = tzgui(ps,parname,gsz,xname,fname)

    if nargin<4
        xname = [];
    end
    if nargin<5
        fname = xname;
    end       
    % parameters    
    psg.par.bcolor  = [0.9 0.9 0.9];        
    psg.par.fcolor  = [1 1 1];             
    psg.par.halign  = 'left';
    if ~isempty(ps)
        psg.par.fn = fieldnames(ps);
    else
        psg.par.fn = [];
    end
    % the grid
    psg.par.gn = gsz;                                    
    % fields
    for i=1:length(psg.par.fn)
        psg.par.p(i).t.text = psg.par.fn{i};           
        psg.par.p(i).p.val  = getfield(ps,psg.par.fn{i});
    end
    
    psg.par.h = gcf;                        % get the current window
    psg.par.sz = get(psg.par.h,'position'); % get its size
    set(psg.par.h,'color',psg.par.bcolor);  % set background color
    set(psg.par.h,'menubar','none');        % no toolbar    
    set(psg.par.h,'numbertitle','off');
    psg.par.sz = psg.par.sz + [0 0 0 2*23];
    set(psg.par.h,'position',psg.par.sz);
    if false
        stck = dbstack;                         % get the caller's name         
        if length(stck)>1
            xname = stck(2).name;                            
        else
            xname = [];
        end
    end
    set(gcf,'name',fname);
        
    psg.par.sz = psg.par.sz([3 4]);
    psg.par.th = 20;                        
    % set the grid of fields
    if psg.par.gn(1)<2
        % grid position increments
        psg.par.dx = (psg.par.sz-[0 psg.par.th])./(psg.par.gn([2 1]));    
        [psg.par.gx psg.par.gy] =  meshgrid(0:psg.par.dx(1):psg.par.sz(1)-psg.par.dx(1),psg.par.sz(2)); 
    else
        % grid position increments
        psg.par.dx = (psg.par.sz-[0 psg.par.th])./(psg.par.gn([2 1])+[0 -1]);    
        [psg.par.gx psg.par.gy] =  meshgrid(0:psg.par.dx(1):psg.par.sz(1)-psg.par.dx(1),psg.par.sz(2):-psg.par.dx(2):0); 
    end
    psg.par.gx = psg.par.gx+2;
    psg.par.gy = psg.par.gy;
    psg.par.xx = psg.par.gx(1);
    psg.par.xy = psg.par.gy(1);
    psg.par.gx = psg.par.gx(2:end);
    psg.par.gy = psg.par.gy(2:end);
    % execute button
    if ~isempty(xname)
        psg.par.xb.pos = [psg.par.xx psg.par.xy-psg.par.th+4 psg.par.dx(1) psg.par.th-2];
        psg.par.xb.h   = uicontrol('Style','pushbutton','string',xname,'position',psg.par.xb.pos,'BackgroundColor','red','HorizontalAlignment','center');
        set(psg.par.xb.h,'fontname','Arial');    
        set(psg.par.xb.h,'Callback',sprintf('GuiExec=1;%s;',xname));
    end
    % parameter fields
    for i=1:min([length(psg.par.fn) length(psg.par.gx)]);
        psg.par.p(i).t.pos = [psg.par.gx(i) psg.par.gy(i)-psg.par.th psg.par.dx(1)/2 psg.par.th];  % a field        
        psg.par.p(i).p.pos = [psg.par.p(i).t.pos(1)+psg.par.p(i).t.pos(3),...
                              psg.par.p(i).t.pos(2),...
                              psg.par.p(i).t.pos(3:4)];  
        psg.par.p(i).t.h   = uicontrol('Style','text','String',psg.par.p(i).t.text,'position',psg.par.p(i).t.pos,'BackgroundColor',psg.par.bcolor,'HorizontalAlignment',psg.par.halign);
        cls = class(psg.par.p(i).p.val);
        switch cls
            case {'double', 'char'}
                set(psg.par.p(i).t.h,'fontname','Arial');
                switch class(psg.par.p(i).p.val)
                    case 'double'
                        txl = num2str(psg.par.p(i).p.val);
                    case 'char'
                        txl = psg.par.p(i).p.val;
                end
                psg.par.p(i).p.h   = uicontrol('Style','edit','string',txl,'position',psg.par.p(i).p.pos,'BackgroundColor',psg.par.bcolor,'HorizontalAlignment',psg.par.halign);        
                set(psg.par.p(i).p.h,'Callback',sprintf('%s=tzsetfield(%s,%.20f,''%s'');',parname,parname,psg.par.p(i).p.h,psg.par.p(i).t.text));
            case 'logical'
                psg.par.p(i).p.h   = uicontrol('Style','radiobutton','position',psg.par.p(i).p.pos,'BackgroundColor',psg.par.bcolor,'HorizontalAlignment',psg.par.halign);        
                set(psg.par.p(i).p.h,'Callback',sprintf('%s=tzsetfield(%s,%.20f,''%s'');',parname,parname,psg.par.p(i).p.h,psg.par.p(i).t.text));
            otherwise
                psg.par.p(i).p.h   = uicontrol('Style','text','position',psg.par.p(i).p.pos,'BackgroundColor',psg.par.bcolor,'HorizontalAlignment',psg.par.halign);                        
                set(psg.par.p(i).p.h,'string',cls);
        end
    end
end
